/**
   MySdlApplication

   Created by Chris Pollett on 2/19/14.

*/
#include <vector>
#include <string>
#include <cmath>
#include <memory>
#include <stdexcept>
#if __GNUG__
#   include <tr1/memory>
#endif

#include "cvec.h"
#include "matrix4.h"
#include "geometrymaker.h"
#include "ppm.h"
#include "glsupport.h"
#include "FragmentApplication.h"

using namespace std;//for string, vector, iostream, and other standard C++ stuff
using namespace tr1;

static const bool g_Gl2Compatible = false;


static const float g_frustMinFov = 60.0;  //A minimal of 60 degree field of view
static float g_frustFovY = g_frustMinFov; // FOV in y direction

static const float g_frustNear = -0.1;    // near plane
static const float g_frustFar = -50.0;    // far plane

static int g_windowWidth = 512;
static int g_windowHeight = 512;
static bool g_mouseClickDown = false;    // is the mouse button pressed
static bool g_mouseLClickButton, g_mouseRClickButton, g_mouseMClickButton;
static int g_mouseClickX, g_mouseClickY; // coordinates for mouse click event
static int g_activeShader = 0;

struct ShaderState {
    GlProgram program;

    // Handles to uniform variables
    GLint h_uProjMatrix;
    GLint h_uModelViewMatrix;
	GLint h_uEyePosition;
	GLint h_uGeometry;
	GLint h_uLights;

    // Handles to vertex attributes
    GLint h_aPosition;


    ShaderState(const char* vsfn, const char* fsfn) {
        readAndCompileShader(program, vsfn, fsfn);

        const GLuint h = program; // short hand

        // Retrieve handles to uniform variables (uniforms can't change per vertex)
        h_uProjMatrix = safe_glGetUniformLocation(h, "uProjMatrix");
        h_uModelViewMatrix = safe_glGetUniformLocation(h, "uModelViewMatrix");
		h_uGeometry = safe_glGetUniformLocation(h, "uGeometry");
		h_uEyePosition = safe_glGetUniformLocation(h, "uEyePosition");
		h_uLights = safe_glGetUniformLocation(h, "uLights");

        // Retrieve handles to vertex attributes (attributes can change per vertex)
        h_aPosition = safe_glGetAttribLocation(h, "aPosition");

        if (!g_Gl2Compatible)
            glBindFragDataLocation(h, 0, "fragColor");
        checkGlErrors();
    }

};

static const int g_numShaders = 2;
static const char * const g_shaderFiles[g_numShaders][2] = {
	{ "./shaders/basic-gl3.vshader", "./shaders/square-test-gl3.fshader" },
    {"./shaders/basic-gl3.vshader", "./shaders/ray-tracer-gl3.fshader"}
};
static const char * const g_shaderFilesGl2[g_numShaders][2] = {
	{ "./shaders/basic-gl2.vshader", "./shaders/square-test-gl2.fshader" },
    {"./shaders/basic-gl2.vshader", "./shaders/ray-tracer-gl2.fshader"}
};
static vector<shared_ptr<ShaderState> > g_shaderStates;
// our global shader states

// --------- Geometry

// Macro used to obtain relative offset of a field within a struct
#define FIELD_OFFSET(StructType, field) &(((StructType *)0)->field)


struct Geometry
{
    GlBufferObject vbo, texVbo, ibo;
    int vboLen, iboLen;

    Geometry(GenericVertex *vtx, unsigned short *idx, int vboLen, int iboLen)
    {
        this->vboLen = vboLen;
        this->iboLen = iboLen;

        // Now create the VBO and IBO
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        glBufferData(GL_ARRAY_BUFFER, sizeof(GenericVertex) * vboLen, vtx,
                     GL_STATIC_DRAW);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned short) * iboLen,
                     idx, GL_STATIC_DRAW);
    }

    void draw(const ShaderState& curSS)
    {
        // Enable the attributes used by our shader
        safe_glEnableVertexAttribArray(curSS.h_aPosition);

        // bind vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        safe_glVertexAttribPointer(curSS.h_aPosition, 3, GL_FLOAT, GL_FALSE,
            sizeof(GenericVertex), FIELD_OFFSET(GenericVertex, pos));
        // bind ibo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);

        // draw!
        glDrawElements(GL_TRIANGLES, iboLen, GL_UNSIGNED_SHORT, 0);

        // Disable the attributes used by our shader
        safe_glDisableVertexAttribArray(curSS.h_aPosition);
    }
};


// Vertex buffer and index buffer associated with the ground and cube geometry
static shared_ptr<Geometry> g_plane;

// --------- Scene

// define light positions in world space
static Matrix4 g_skyRbt = Matrix4::makeTranslation(Cvec3(0.0, 0.0, 1.5));
static Matrix4 g_objectRbt[1] = {Matrix4::makeTranslation(Cvec3(0,0,0))};
// currently only 1 obj is defined
static Cvec3f g_objectColors[1] = {Cvec3f(1, 0, 0)};

/*
   Data that will be sent to the shader for ray tracing
   Format in array for shapes:
   If type is:
   TETRAHEDRON: type, edge_length, c_x, c_y, c_z, dummy
   CUBE: type, edge_length, c_x, c_y, c_z, dummy
   SPHERE: type, radius, c_x, c_y, c_z, dummy
   CYLINDER: type, radius, height, c_x, c_y, c_z
   CONE: type, radius, height, c_x, c_y, c_z
 */
static GLfloat g_eyePosition[3] = { 0.0, 0.0, 1.0};
#define NUM_LIGHTS 1
#define LIGHT_STRIDE 3
static GLfloat g_Lights[3] = { 0.0, 5.0, 0.0 };
#define TETRAHEDRON 0
#define CUBE 1
#define SPHERE 2
#define CYLINDER 3
#define CONE 4
#define NUM_SHAPES 1
#define GEOMETRY_STRIDE 6
//for your code you get this geometry data from user input
static GLfloat g_geometryData[NUM_SHAPES * GEOMETRY_STRIDE] = 
	{ SPHERE, -0.2, 0.0, 0.0, 0.0, 0.0 };

///////////////// END OF G L O B A L S ///////////////////////


static void initPlane()
{
    int ibLen, vbLen;
    getPlaneVbIbLen(vbLen, ibLen);

    // Temporary storage for cube geometry
    vector<GenericVertex> vtx(vbLen);
    vector<unsigned short> idx(ibLen);

    makePlane(1, vtx.begin(), idx.begin());
    g_plane.reset(new Geometry(&vtx[0], &idx[0], vbLen, ibLen));
}

// takes a projection matrix and send to the the shaders
static void sendProjectionMatrix(const ShaderState& curSS,
                                 const Matrix4& projMatrix)
{
    GLfloat glmatrix[16];
    projMatrix.writeToColumnMajorMatrix(glmatrix); // send projection matrix
    safe_glUniformMatrix4fv(curSS.h_uProjMatrix, glmatrix);
}

// takes MVM and its normal matrix to the shaders
static void sendGeometry(const ShaderState& curSS,
                                      const Matrix4& MVM)
{
    GLfloat glmatrix[16];
    MVM.writeToColumnMajorMatrix(glmatrix); // send MVM
    safe_glUniformMatrix4fv(curSS.h_uModelViewMatrix, glmatrix);

	glUniform1fv(curSS.h_uEyePosition, 3, g_eyePosition);
	glUniform1fv(curSS.h_uLights, NUM_LIGHTS * LIGHT_STRIDE, g_geometryData);
	glUniform1fv(curSS.h_uGeometry, NUM_SHAPES * GEOMETRY_STRIDE, g_geometryData);
}

// update g_frustFovY from g_frustMinFov, g_windowWidth, and g_windowHeight
static void updateFrustFovY()
{
    if (g_windowWidth >= g_windowHeight)
        g_frustFovY = g_frustMinFov;
    else {
        const double RAD_PER_DEG = 0.5 * CS175_PI/180;
        g_frustFovY = atan2(sin(g_frustMinFov * RAD_PER_DEG) * g_windowHeight
                            / g_windowWidth,
                            cos(g_frustMinFov * RAD_PER_DEG)) / RAD_PER_DEG;
    }
}

static Matrix4 makeProjectionMatrix()
{
    return Matrix4::makeProjection(g_frustFovY,
        g_windowWidth / static_cast <double> (g_windowHeight),
        g_frustNear, g_frustFar);
}

static void drawStuff()
{
    // short hand for current shader state
    const ShaderState& curSS = *g_shaderStates[g_activeShader];

    // build & send proj. matrix to vshader
    const Matrix4 projmat = makeProjectionMatrix();
    sendProjectionMatrix(curSS, projmat);

    // use the skyRbt as the eyeRbt
    const Matrix4 eyeRbt = g_skyRbt;
    const Matrix4 invEyeRbt = inv(eyeRbt);

    const Matrix4 groundRbt = Matrix4();  // identity
    Matrix4 MVM = invEyeRbt * groundRbt;
    // draw plan
    // ==========
    MVM = invEyeRbt * g_objectRbt[0];
    sendGeometry(curSS, MVM);
    g_plane->draw(curSS);
}


static void initGLState()
{
    glClearColor(128./255., 200./255., 255./255., 0.);
    glClearDepth(0.);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_GREATER);
    glReadBuffer(GL_BACK);
    if (!g_Gl2Compatible) {
        glEnable(GL_FRAMEBUFFER_SRGB);
    }
}

static void initShaders()
{
    g_shaderStates.resize(g_numShaders);
    for (int i = 0; i < g_numShaders; ++i) {
        if (g_Gl2Compatible) {
            g_shaderStates[i].reset(new ShaderState(g_shaderFilesGl2[i][0],
                                                    g_shaderFilesGl2[i][1]));
        } else {
            g_shaderStates[i].reset(new ShaderState(g_shaderFiles[i][0],
                                                    g_shaderFiles[i][1]));
        }
    }
}

static void initGeometry()
{
    initPlane();
}


FragmentApplication::FragmentApplication() {
    running = true;
}

bool FragmentApplication::onInit() {
    if(SDL_Init(SDL_INIT_EVERYTHING) < 0) {
        return false;
    }

    /* Turn on double buffering with a 24bit Z buffer.
     * You may need to change this to 16 or 32 for your system */
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    if((display = SDL_CreateWindow("Ray Tracing Fragment Shader",
        SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 640, 480,
        SDL_WINDOW_OPENGL)) == NULL) {
        return false;
    }

    /* Create our opengl context and attach it to our window */

    SDL_GLContext maincontext = SDL_GL_CreateContext(display);
    /* This makes our buffer swap syncronized with the 
      monitor's vertical refresh */
    SDL_GL_SetSwapInterval(1);

    glewInit();
    cout << (g_Gl2Compatible ?
        "Will use OpenGL 2.x / GLSL 1.0" : "Will use OpenGL 3.x / GLSL 1.3")
        << endl;
    if ((!g_Gl2Compatible) && !GLEW_VERSION_3_0)
        throw runtime_error(
            "Error: does not support OpenGL Shading Language v1.3");
    else if (g_Gl2Compatible && !GLEW_VERSION_2_0)
        throw runtime_error(
            "Error: does not support OpenGL Shading Language v1.0");

    initGLState();
    initShaders();
    initGeometry();

    return true;
}

void FragmentApplication::onEvent(SDL_Event* event) {
    switch(event->type)
    {
        case SDL_QUIT:
            running = false;
            break;
        case SDL_KEYUP:
            keyboard(SDL_GetKeyName(event->key.keysym.sym));
            break;
        case SDL_MOUSEBUTTONDOWN:
            mouse(event->button);
            break;
        case SDL_MOUSEBUTTONUP:
            mouse(event->button);
            break;
        case SDL_MOUSEMOTION:
            motion(event->motion.x, event->motion.y);
            break;
    }
}

void FragmentApplication::keyboard(const char* key)
{
    cout << key << endl;
    switch (key[0])
    {
        case 27: //ESC
            exit(0);
        case 'H':
            cout << " ============== H E L P ==============\n\n"
            << "h\t\thelp menu\n"
            << "s\t\tsave screenshot\n"
            << "t\t\ttoggle between shaders\n"
            << "drag left mouse to rotate\n" << endl;
            break;
        case 'S':
            glFlush();
            writePpmScreenshot(g_windowWidth, g_windowHeight, "out.ppm");
            break;
        case 'T':
            g_activeShader++;
            if(g_activeShader >= g_numShaders) {
                g_activeShader = 0;
            }
            break;
    }
}

void FragmentApplication::mouse(SDL_MouseButtonEvent button)
{
    g_mouseClickX = button.x;
    g_mouseClickY = g_windowHeight - button.y - 1;

    g_mouseLClickButton |= (button.button == SDL_BUTTON_LEFT &&
                            button.state == SDL_PRESSED);
    g_mouseRClickButton |= (button.button == SDL_BUTTON_RIGHT &&
                            button.state == SDL_PRESSED);
    g_mouseMClickButton |= (button.button == SDL_BUTTON_MIDDLE &&
                            button.state == SDL_PRESSED);

    g_mouseLClickButton &= !(button.button == SDL_BUTTON_LEFT &&
                            button.state == SDL_RELEASED);
    g_mouseRClickButton &= !(button.button == SDL_BUTTON_RIGHT &&
                             button.state == SDL_RELEASED);
    g_mouseMClickButton &= !(button.button == SDL_BUTTON_MIDDLE &&
                             button.state == SDL_RELEASED);

    g_mouseClickDown = g_mouseLClickButton || g_mouseRClickButton ||
        g_mouseMClickButton;
}

void FragmentApplication::motion(const int x, const int y)
{
    const double dx = x - g_mouseClickX;
    const double dy = g_windowHeight - y - 1 - g_mouseClickY;

    Matrix4 m;
    if (g_mouseLClickButton && !g_mouseRClickButton) {
        // left button down?
        m = Matrix4::makeXRotation(-dy) * Matrix4::makeYRotation(dx);
    } else if (g_mouseRClickButton && !g_mouseLClickButton) {
        // right button down?
        m = Matrix4::makeTranslation(Cvec3(dx, dy, 0) * 0.01);
    } else if (g_mouseMClickButton ||
               (g_mouseLClickButton && g_mouseRClickButton)) {
        // middle or (left and right) button down?
        m = Matrix4::makeTranslation(Cvec3(0, 0, -dy) * 0.01);
    }

    if (g_mouseClickDown) {
        g_objectRbt[0] *= m; // Simply right-multiply is WRONG
    }

    g_mouseClickX = x;
    g_mouseClickY = g_windowHeight - y - 1;
}

void FragmentApplication::onLoop() {
}

void FragmentApplication::onRender() {
    glUseProgram(g_shaderStates[g_activeShader]->program);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    // clear framebuffer color&depth
    drawStuff();
    
    SDL_GL_SwapWindow(display);
    checkGlErrors();
}

void FragmentApplication::onCleanup() {
    SDL_Quit();
}

int FragmentApplication::onExecute() {
    if(onInit() == false) {
        return -1;
    }
    SDL_Event Event;
    while(running) {
        while(SDL_PollEvent(&Event)) {
            onEvent(&Event);
        }

        onLoop();
        onRender();
    }

    onCleanup();

    return 0;
}

int main(int argc, const char * argv[])
{
    FragmentApplication application;
    return application.onExecute();
}

